# -*- coding: utf-8 -*-

import os

import psutil

from typing import Optional

from PyQt5.QtCore import QDir, QCoreApplication, QProcess
from PyQt5.QtGui import QIcon

from libastra_csp.common_functions import ResultStatus, check_result, emergency_exit
from libastra_csp.container_creation_thread import ContainerCreationThread
from libastra_csp.package import Package
from libfly_csp.base_main_window import BaseMainWindow
from libfly_csp.wait_thread_dialog import WaitThreadDialog
from astra_csp_cryptopro.certificate_process import CertificateProcess
from astra_csp_cryptopro.create_file_signature_process import CreateFileSignatureProcess
from astra_csp_cryptopro.merge_certificates_process import MergeCertificatesProcess
from astra_csp_cryptopro.verify_file_process import VerifyFileProcess
from astra_csp_cryptopro.cryptopro_settings import cryptopro_settings
from astra_csp_cryptopro.common_functions import tuned_merge_certificate_process, \
    get_status, tuned_certificates_process, tuned_verify_file_process, \
    tuned_file_signature_process, validate_system, setup_cryptopro_bin_dir, remove_write_permissions_handler, \
    get_enriched_output, validate_cryptopro_config
from fly_csp_cryptopro.cprocsp_process_dialog import CprocspProcessDialog
from fly_csp_cryptopro.graphic_cryptopro_settings import graphic_cryptopro_settings

translate = QCoreApplication.translate


class MainWindow(BaseMainWindow):
    def __init__(self) -> None:
        super().__init__()
        self.cryptopro_bin_dir = None  # type: Optional[QDir]
        self.cached_time_mark_plugin_installed_value = None  # type: Optional[bool]

        self.setWindowIcon(QIcon.fromTheme('fly-csp-cryptopro'))

    def initialize(self) -> None:
        super().initialize()

        try:
            validate_system()
            validate_cryptopro_config()
            self.validate_required_packages()
            self.cryptopro_bin_dir = setup_cryptopro_bin_dir()
            self.cached_time_mark_plugin_installed_value = self.time_mark_plugin_installed()

            if self.cached_time_mark_plugin_installed_value:
                self.checkBoxTimestampCheck.setToolTip(
                    translate('text', 'If the signature was created with a time stamp, '
                                      'then you need to activate this checkbox'))
                self.checkBoxIncludeTimestamp.stateChanged.connect(self.on_check_box_include_timestamp_state_changed)
            else:
                self.checkBoxTimestampCheck.setEnabled(False)
                self.checkBoxTimestampCheck.setChecked(False)

                list_of_packages = cryptopro_settings.time_mark_plugins or []
                text = ', '.join(translate('text', ' or ').join(packages) for packages in list_of_packages or [])
                timestamp_tooltip_text = translate('text',
                                                   'Package(s) {text} not installed',
                                                   None,
                                                   len(list_of_packages)).format(text=text)
                self.checkBoxTimestampCheck.setToolTip(timestamp_tooltip_text)

                self.checkBoxIncludeTimestamp.setEnabled(False)
                self.checkBoxIncludeTimestamp.setChecked(False)
                self.checkBoxIncludeTimestamp.setToolTip(timestamp_tooltip_text)
        except Exception as exception:
            self.show_critical_error(translate('text', 'Critical error'),
                                     translate('text', 'Failed to initialize system'),
                                     str(exception))

    def start(self) -> None:
        super().start()

        merge_certificates_process = tuned_merge_certificate_process(
            callback=self.on_merge_certificates_process_completed)

        merge_wait_dialog = CprocspProcessDialog(process=merge_certificates_process, parent=self)
        self.start_waiting(merge_wait_dialog)
        merge_wait_dialog.setWindowTitle(translate('text', 'Adding user certificates to certificate store'))
        merge_certificates_process.start()

    def on_merge_certificates_process_completed(self) -> None:
        process = self.sender()  # type: MergeCertificatesProcess
        if type(process) != MergeCertificatesProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            emergency_exit(self.cache)

        try:
            check_result(process)
        except Exception as exception:
            self.show_error(translate('text', 'Error'),
                            translate('text', 'Failed to add user certificates to certificate store'),
                            ' '.join(exception.args))

        self.enable_ui()

        certificates_process = tuned_certificates_process(
            callback=self.on_get_certificates_process_completed,
            kill_time=self.common_settings.max_csp_time_waiting or 0)

        get_certificates_dialog = CprocspProcessDialog(process=certificates_process, parent=self)
        self.start_waiting(get_certificates_dialog)
        get_certificates_dialog.setWindowTitle(translate('text', 'Loading certificates'))
        certificates_process.start()

    def on_get_certificates_process_completed(self) -> None:
        process = self.sender()  # type: CertificateProcess
        if type(process) != CertificateProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            emergency_exit(self.cache)

        try:
            check_result(process)
        except Exception as exception:
            self.show_critical_error(translate('text', 'Critical error'),
                                     translate('text', 'Failed to load certificates'), ' '.join(exception.args))
            return

        if process.empty_list_error_occurred():
            self.show_critical_error(translate('text', 'Failed to load certificates'),
                                     translate('text', 'Certificates not available. Empty certificate list.'))
            return

        try:
            certificates = process.get_certificates()
        except Exception as exception:
            self.show_critical_error(translate('text', 'Failed to load certificates'),
                                     str(exception))
            return

        if not certificates:
            self.show_critical_error(translate('text', 'Failed to load certificates'),
                                     translate('text', 'Certificates not available. Empty certificate list.'))

        self.setup_certificates(certificates)

        self.enable_ui()
        self.maybe_verify()

    def on_create_file_signature_process_completed(self) -> None:
        process = self.sender()  # type: CreateFileSignatureProcess
        if type(process) != CreateFileSignatureProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            self.clean_cancelled_signature(process.signature)
            return

        try:
            check_result(process, check_exit_code=False)
        except Exception as exception:
            self.show_error(translate('text', 'Error'),
                            translate('text', 'Failed to create file signature'), ' '.join(exception.args))

        status, output = remove_write_permissions_handler(process)

        self.show_signing_result(output, status)

        if status == ResultStatus.Error:
            return
        elif self.checkBoxCreateContainer.isChecked():
            self.create_container(process.document,
                                  process.signature)
        elif self.container:
            container_creation_thread = ContainerCreationThread(container=self.container,
                                                                files=[process.document, process.signature],
                                                                update=True)
            container_creation_thread.finished.connect(self.on_update_container_thread_finished)
            wait_dialog = WaitThreadDialog(thread=container_creation_thread, parent=self)
            self.start_waiting(wait_dialog)
            container_creation_thread.start()
        else:
            self.change_signature_file(process.signature)

    def on_verify_file_process_completed(self) -> None:
        process = self.sender()  # type: VerifyFileProcess
        if type(process) != VerifyFileProcess:
            return

        is_cancelled = self.waiting_dialog.cancelled
        self.stop_waiting()
        if is_cancelled:
            return

        try:
            check_result(process, check_exit_code=False)
        except Exception as exception:
            self.show_error(translate('text', 'Error'), translate('text', 'Failed to verify file'),
                            ' '.join(exception.args))
            return

        status = get_status(process)

        self.show_verification_result(get_enriched_output(process), status)

    def enable_signature_actions(self) -> None:
        super().enable_signature_actions()
        self.checkBoxTimestampCheck.setEnabled(bool(self.cached_time_mark_plugin_installed_value))

    def enable_controls(self) -> None:
        super().enable_controls()
        if self.cached_time_mark_plugin_installed_value:
            self.checkBoxIncludeTimestamp.setEnabled(True)

    def verify_document(self) -> None:
        super().verify_document()

        verify_file_process = tuned_verify_file_process(document=self.document,
                                                        sig_path=self.signature,
                                                        check_timestamp=self.check_timestamp(),
                                                        callback=self.on_verify_file_process_completed)
        self.start_waiting(CprocspProcessDialog(process=verify_file_process, parent=self))
        verify_file_process.start()

    def sign_document(self) -> None:
        super().sign_document()

        dirname = os.path.dirname(self.document)
        filename = self.lineEditSignatureFileName.text()
        filepath = "{dir}/{file}".format(dir=dirname, file=filename)
        fprint = self.selected_certificate.fingerprint

        create_file_signature_process = tuned_file_signature_process(
            document=self.document,
            sig_name=filepath,
            fingerprint=fprint,
            der_format=self.comboBoxFormat.currentText() == 'DER',
            time_address=cryptopro_settings.time_address
            if self.include_timestamp() else None,
            callback=self.on_create_file_signature_process_completed)

        self.start_waiting(CprocspProcessDialog(process=create_file_signature_process, parent=self))
        create_file_signature_process.start()

    def change_signature_file(self, file_path: Optional[str]) -> None:
        super().change_signature_file(file_path)

        if self.signature_contains_timestamp_extension():
            if not self.cached_time_mark_plugin_installed_value:
                self.checkBoxTimestampCheck.setChecked(False)
                list_of_packages = cryptopro_settings.time_mark_plugins or []
                text = ', '.join(translate('text', ' or ').join(packages) for packages in list_of_packages or [])
                self.show_warning(translate('text', 'Warning'),
                                  translate('text',
                                            'Timestamped signature file checking not possible. '
                                            'Package(s) {text} not installed. '
                                            'An file signature check will be performed without a timestamp.',
                                            None,
                                            len(list_of_packages)).format(text=text))
            else:
                self.checkBoxTimestampCheck.setChecked(True)
        else:
            self.checkBoxTimestampCheck.setChecked(False)

    def signature_contains_timestamp_extension(self) -> bool:
        if not self.signature:
            return False

        basename = os.path.basename(self.signature)
        parts = basename.split('.')
        if len(parts) > 1:
            return 'tsp' in parts[1:]  # TODO: move to config
        else:
            return False

    def show_waiting_dialog(self) -> None:
        if not self.is_cryptopro_ui_running() and self.waiting_dialog:
            self.waiting_dialog.exec()

    def show_help(self) -> None:
        url = 'qthelp://flyhelp.ru/doc/fly-csp-cryptopro/fly-csp-cryptopro.html'
        path = '/usr/share/doc/fly/qthelp/ru/flyhelp.qhc'
        QProcess.startDetached("assistant -collectionFile {path} -showUrl {url}".format(path=path, url=url))

    @staticmethod
    def is_cryptopro_ui_running() -> bool:
        uid = os.getuid()
        for proc in psutil.process_iter():
            try:
                if proc.name() == "xcpui_app" and proc.uids().real == uid:
                    return True
            except psutil.Error:
                pass
        return False

    @staticmethod
    def time_mark_plugin_installed() -> bool:
        for packages in cryptopro_settings.time_mark_plugins or []:
            if not any(Package.installed(package) for package in packages):
                return False
        else:
            return True

    @staticmethod
    def validate_required_packages() -> None:
        all_not_installed_packages = []

        for packages in graphic_cryptopro_settings.required_gui_components_packages or []:
            not_installed = []
            for package in packages:
                if not Package.installed(package):
                    not_installed.append(package)
            if not_installed:
                all_not_installed_packages.append(not_installed)

        if all_not_installed_packages:
            packages = ['"{text}"'.format(text=translate('text', ' or ').join(packages)) for packages
                        in all_not_installed_packages]
            raise Exception(translate('text',
                                      'Package {packages} not installed',
                                      None,
                                      len(packages))
                            .format(packages=translate('text', ' and ').join(packages)))
